import torch
import torch.nn as nn


def threshold_weight_3sd(module, qmodule, eps=1e-8):
    mean_value = torch.tensor([module.weight.mean().abs().data])
    std_value = module.weight.std().data
    qmodule.weight_log2_t = torch.nn.Parameter(
        torch.log2(mean_value +
                   3 * std_value)) if qmodule.retrain else torch.log2(
                       mean_value + 3 * std_value + eps)


def threshold_bias_3sd(module, qmodule, eps=1e-8):
    mean_value = torch.tensor([module.bias.mean().abs().data])
    std_value = module.bias.std().data
    qmodule.bias_log2_t = torch.nn.Parameter(
        torch.log2(mean_value +
                   3 * std_value)) if qmodule.retrain else torch.log2(
                       mean_value + 3 * std_value + eps)


def threshold_activation_3sd(module, qmodule, eps=1e-8):
    mean_value = torch.tensor([module.hook_out.mean().abs().data])
    std_value = module.hook_out.std().data
    qmodule.acti_log2_t = torch.nn.Parameter(
        torch.log2(mean_value +
                   3 * std_value)) if qmodule.retrain else torch.log2(
                       mean_value + 3 * std_value + eps)



import torch
import torch.nn as nn


def threshold_weight_max(module, qmodule, eps=1e-8):
    max_value = torch.tensor([module.weight.abs().flatten().max().data]) + eps
    qmodule.weight_log2_t = torch.nn.Parameter(
        torch.log2(max_value)) if qmodule.retrain else torch.log2(max_value)


def threshold_bias_max(module, qmodule, eps=1e-8):
    max_value = torch.tensor([module.bias.abs().flatten().max().data]) + eps
    qmodule.bias_log2_t = torch.nn.Parameter(
        torch.log2(max_value)) if qmodule.retrain else torch.log2(max_value)


def threshold_activation_max(module, qmodule, eps=1e-8):
    max_value = torch.tensor([module.hook_out.abs().flatten().max().data
                              ]) + eps
    qmodule.acti_log2_t = torch.nn.Parameter(
        torch.log2(max_value)) if qmodule.retrain else torch.log2(max_value)



import torch
import torch.nn as nn
import math


def kl_divergence(ha, hb):
    r'''
        ans = J_{kl}(a,b)
    '''
    return torch.sum(ha * torch.log(ha / hb))


def quantize_bins_and_expand(dist, quant_bins):
    dist_len = dist.shape[0]
    width = math.floor(1. * dist_len / quant_bins)
    dist_q = torch.zeros([quant_bins])
    dist_e = 0. * dist
    for i in range(quant_bins):
        if i != quant_bins - 1:
            dist_q[i] = dist[i * width:(i + 1) * width].sum()
            width_preserve = width - (dist[i * width:(i + 1) * width]
                                      == 0).sum()
            dist_e[i * width:(i + 1) * width] = dist_q[i] / width_preserve * (
                dist[i * width:(i + 1) * width] == 0)
        else:
            dist_q[i] = dist[i * width:].sum()
            width_preserve = width - (dist[i * width:] == 0).sum()
            dist_e[i * width:] = dist_q[i] / width_preserve * (dist[i * width:]
                                                               == 0)
    return dist_e


def entropy_calibration(model,
                        qmodel,
                        bin_number=2048,
                        cali_number=128,
                        eps=1e-8):
    q = model.hook_out.flatten().data
    dist = torch.histc(q, bins=bin_number)
    bin_width = (q.max() - q.min()) / bin_number
    divergence = torch.zeros([bin_number]) * 1.0
    for i in range(cali_number, bin_number):
        ref_dist = dist[:i]
        outliers_count = dist[i:].sum()
        ref_dist[-1] += outliers_count
        ref_dist /= ref_dist.sum()
        can_dist = quantize_bins_and_expand(dist[:i], cali_number)
        can_dist /= can_dist.sum()
        divergence[i] = kl_divergence(ref_dist, can_dist)

    m, m_idx = torch.min(divergence[cali_number:], 0)
    threshold = q.min() + (m_idx + cali_number + 0.5) * bin_width + eps
    log2_t = torch.tensor([torch.log2(threshold)])
    qmodel.acti_log2_t = torch.nn.Parameter(
        log2_t) if qmodel.retrain else log2_t



"""
Provide quantilized form of torch.nn.modules.batchnorm 
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from .number import qsigned


class BatchNorm2d(nn.BatchNorm2d):
    def __init__(self,
                 num_features,
                 eps=1e-5,
                 momentum=0.1,
                 affine=True,
                 track_running_stats=True,
                 weight_bit_width=8,
                 bias_bit_width=16,
                 retrain=True,
                 quant=False):
        super().__init__(num_features, eps, momentum, affine,
                         track_running_stats)

        self.weight_bit_width = weight_bit_width
        self.bias_bit_width = bias_bit_width
        self.retrain = retrain
        self.quant = quant
        if retrain is True:
            self.weight_log2_t = nn.Parameter(torch.Tensor(1))
            if self.bias is not None:
                self.bias_log2_t = nn.Parameter(torch.Tensor(1))
        else:
            self.weight_log2_t = torch.Tensor(1)
            if self.bias is not None:
                self.bias_log2_t = torch.Tensor(1)

    def static(self):
        self.retrain = False
        if isinstance(self.bias_log2_t, nn.Parameter):
            self.bias_log2_t.requires_grad_(False)
        if isinstance(self.weight_log2_t, nn.Parameter):
            self.weight_log2_t.requires_grad_(False)

    def quantilize(self):
        self.quant = True

    def floatilize(self):
        self.quant = False

    def bn_forward(self, input):
        if self.affine is True:
            weight = qsigned(self.weight, self.weight_log2_t,
                             self.weight_bit_width)
            bias = qsigned(self.bias, self.bias_log2_t, self.bias_bit_width)
            output = F.batch_norm(input,
                                  running_mean=self.running_mean,
                                  running_var=self.running_var,
                                  weight=weight,
                                  bias=bias)
        else:
            output = F.batch_norm(input,
                                  running_mean=self.running_mean,
                                  running_var=self.running_var)
        return output

    def bn_forward_unquant(self, input):
        if self.affine is True:
            output = F.batch_norm(input,
                                  running_mean=self.running_mean,
                                  running_var=self.running_var,
                                  weight=self.weight,
                                  bias=self.bias)
        else:
            output = F.batch_norm(input,
                                  running_mean=self.running_mean,
                                  running_var=self.running_var)
        return output

    def forward(self, input):
        return self.bn_forward(
            input) if self.quant else self.bn_forward_unquant(input)



"""
Provide Adder2d, https://arxiv.org/pdf/1912.13200.pdf
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Function
import math


class Adder(Function):
    @staticmethod
    def forward(ctx, W_col, X_col):
        ctx.save_for_backward(W_col, X_col)
        output = -(W_col.unsqueeze(2) - X_col.unsqueeze(0)).abs().sum(1)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        W_col, X_col = ctx.saved_tensors
        grad_W_col = ((X_col.unsqueeze(0) - W_col.unsqueeze(2)) *
                      grad_output.unsqueeze(1)).sum(2)
        grad_W_col = grad_W_col / grad_W_col.norm(p=2).clamp(
            min=1e-12) * math.sqrt(W_col.size(1) * W_col.size(0)) / 5
        grad_X_col = (-(X_col.unsqueeze(0) - W_col.unsqueeze(2)).clamp(-1, 1) *
                      grad_output.unsqueeze(1)).sum(0)

        return grad_W_col, grad_X_col


adder = Adder.apply


def adder2d_function(X, W, bias, stride=1, padding=0):
    n_filters, d_filter, h_filter, w_filter = W.size()
    n_x, d_x, h_x, w_x = X.size()

    h_out = (h_x - h_filter + 2 * padding) / stride + 1
    w_out = (w_x - w_filter + 2 * padding) / stride + 1

    h_out, w_out = int(h_out), int(w_out)
    X_col = torch.nn.functional.unfold(X.view(1, -1, h_x, w_x),
                                       h_filter,
                                       dilation=1,
                                       padding=padding,
                                       stride=stride).view(
                                           n_x, -1, h_out * w_out)
    X_col = X_col.permute(1, 2, 0).contiguous().view(X_col.size(1), -1)
    W_col = W.view(n_filters, -1)

    out = adder(W_col, X_col)

    out = out.view(n_filters, h_out, w_out, n_x)
    out = out.permute(3, 0, 1, 2).contiguous()

    if bias is not None:
        out += bias.unsqueeze(0).unsqueeze(2).unsqueeze(3)

    return out


class Adder2d(nn.Module):
    def __init__(self,
                 input_channel,
                 output_channel,
                 kernel_size,
                 stride=1,
                 padding=0,
                 bias=False):
        super().__init__()
        self.stride = stride
        self.padding = padding
        self.input_channel = input_channel
        self.output_channel = output_channel
        self.kernel_size = kernel_size
        self.weight = torch.nn.Parameter(
            nn.init.normal_(
                torch.randn(output_channel, input_channel, kernel_size,
                            kernel_size)))
        self._bias = bias
        if bias:
            self.bias = torch.nn.Parameter(
                nn.init.uniform_(torch.zeros(output_channel)))
        else:
            self.bias = None

    def forward(self, x):
        output = adder2d_function(x, self.weight, self.bias, self.stride,
                                  self.padding)
        return output


if __name__ == '__main__':
    add = Adder2d(3, 4, 3, bias=True)
    x = torch.rand(10, 3, 10, 10)
    print(add(x).shape)



"""
Provide quantilized form of Adder2d, https://arxiv.org/pdf/1912.13200.pdf
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Function

from . import extra as ex
from .number import qsigned


class Adder2d(ex.Adder2d):
    def __init__(self,
                 input_channel,
                 output_channel,
                 kernel_size,
                 stride=1,
                 padding=0,
                 bias=False,
                 weight_bit_width=8,
                 bias_bit_width=16,
                 retrain=True,
                 quant=False):
        super().__init__(input_channel,
                         output_channel,
                         kernel_size,
                         stride=stride,
                         padding=padding,
                         bias=bias)
        self.weight_bit_width = weight_bit_width
        self.bias_bit_width = bias_bit_width
        self.retrain = retrain
        self.quant = quant
        if retrain is True:
            self.weight_log2_t = nn.Parameter(torch.Tensor(1))
            if self.bias is not None:
                self.bias_log2_t = nn.Parameter(torch.Tensor(1))
        else:
            self.weight_log2_t = torch.Tensor(1)
            if self.bias is not None:
                self.bias_log2_t = torch.Tensor(1)
        pass

    def quantilize(self):
        self.quant = True

    def floatilize(self):
        self.quant = False

    def adder_forward(self, input):
        if self.bias is None:
            bias = None
        else:
            bias = qsigned(self.bias, self.bias_log2_t, self.bias_bit_width)
        weight = qsigned(self.weight, self.weight_log2_t,
                         self.weight_bit_width)
        return ex.adder2d_function(input,
                                   weight,
                                   bias,
                                   stride=self.stride,
                                   padding=self.padding)

    def adder_forward_unquant(self, input):
        return ex.adder2d_function(input,
                                   self.weight,
                                   self.bias,
                                   stride=self.stride,
                                   padding=self.padding)

    def forward(self, input):
        return self.adder_forward(
            input) if self.quant else self.adder_forward_unquant(input)


if __name__ == '__main__':
    add = Adder2d(3, 4, 3, bias=True)
    x = torch.rand(10, 3, 10, 10)
    print(add(x).shape)
